; 'Remote Alarm' V1.02 Designed by Myriad Development Co. - Jerry Farmer
;       PIC16C54, 4MHz Crystal, WatchDog Timer OFF, MPASM instruction set

        list    p=16C54,t=ON,c=132

        include "PICREG.H"

; Constants
INDIR           equ     0               ;Indirect Register
OUT_BIT_TIME    equ     33h             ;9600 Baud, 104uS Bit Rate
IN_BIT_TIME     equ     64h             ;4800 Baud, 208uS Bit Rate
FUDGE_TIME      equ     23h            ;Current Time within a Fudge Factor

; B Register Definitions
#define Level_Reset     Port_B,0        ;Low will cause Past Level to reset
                                        ;RB.7 - RB.1 == Input from Sensors
RB_TRIS         equ     B'11111111'     ;RB TRIS at INIT State == all input
RB_MASK         equ     B'00000000'     ;What is High/Low for RB at INIT State

; A Register Definitions - Programmable Inputs
#define Serial_IN_1     Port_A,0        ;Serial Input #1 - 8 bits
#define LED             Port_A,1        ;LED Output - Level/State Indicator
#define Serial_Out      Port_A,2        ;Serial Output - 8 bits + passwords
#define Serial_IN_2     Port_A,3        ;Serial Input #2 - 8 bits

RA_TRIS         equ     B'11111001'     ;RA TRIS at INIT State
RA_MASK         equ     B'00000000'     ;What is High/Low for RA at INIT State

; Register Files
temp            equ     07h             ;Tempary holding register - PIC16C54/56
Timer_Bits      equ     08h             ;Indicates which Timer(s) are Active = 1
Flags           equ     09h             ;Error Flags
LED_Mode        equ     0Ah             ;(0-2)=Mode, 3=LED_B, (4-6)=Seq #, 7=NEW
OState          equ     0Bh             ;Serial Out State
T_5_M_LO        equ     0Ch             ;5 Min Timer Counter - Low
T_5_M_HI        equ     0Dh             ;5 Min Timer Counter - High
T_5_S_CO        equ     0Eh             ;5 Second Timer - lack of Serial Input
T_20_mS_CO      equ     0Fh             ;20 mS Timer - used for debouncing
LED_C           equ     10h             ;LED Counter
Last_RTCC       equ     11h             ;Last value of the RTCC
First_RTCC_O    equ     12h             ;Starting time for next Output event
xmt_byte        equ     13h             ;Serial xmit byte - destroyed in use
cc              equ     14h             ;256 * RTCC time
RCV_Storage     equ     15h             ;Long term storage of rcv_byte #1 & 2
Old_RB          equ     16h             ;Oldest/Master copy of RB
Last_RB         equ     17h             ;Last copy of RB
IState1         equ     18h             ;Serial In #1 State
First_RTCC_I1   equ     19h             ;Starting time for next #1 Input event
nbti1           equ     1Ah             ;Next Bit #1 In Time - variable time 
rcv_byte_1      equ     1Bh             ;Receive Serial #1 In byte
IState2         equ     1Ch             ;Serial In #2 State
First_RTCC_I2   equ     1Dh             ;Starting time for next #2 Input event
nbti2           equ     1Eh             ;Next Bit #2 In Time - variable time 
rcv_byte_2      equ     1Fh             ;Receive Serial #2 In byte

; Indicates which Timer(s) are Active = 1 & Flags
#define OState_B        Timer_Bits,0    ;Serial Out Active Bit
#define IState1_B       Timer_Bits,1    ;Serial IN #1 Active Bit
#define IState2_B       Timer_Bits,2    ;Serial IN #2 Active Bit
#define T_5_S_B         Timer_Bits,3    ;5 Second Timer Active Bit
#define T_5_M_B         Timer_Bits,4    ;5 Min Timer Active Bit
#define RCV_Got_One_B   Timer_Bits,5    ;Got a NEW Received byte to send out
#define RB_NEW_B        Timer_Bits,6    ;Indicates a change in RB input
#define S_5_S_B         Timer_Bits,7    ;Serial In 5 secs of inactivity

; Error Flags
#define FS_Flag_1       Flags,0         ;Serial #1 IN had a False Start Error
#define FE_Flag_1       Flags,1         ;Last Serial #1 IN had a Frame Error
#define FS_Flag_2       Flags,2         ;Serial #2 IN had a False Start Error
#define FE_Flag_2       Flags,3         ;Last Serial #2 IN had a Frame Error
#define RCV_Overflow    Flags,4         ;Lost Serial Input Byte - too Slow
#define RB_Overflow     Flags,5         ;Lost RB Input Byte - too Slow
#define S_5_S_Overflow  Flags,6         ;Lost '5S Inactivity' msg - too Slow

;Equates for LED Task #7
#define LED_B           LED_Mode,3      ;LED is active
#define LED_NEW_B       LED_Mode,7      ;LED has just changed Modes = 1
LED_OFF_MODE    equ     B'00001000'     ;LED OFF
LED_SEQ1_MODE   equ     B'10001001'     ;LED Sequence 1: .2s On, 1s Off
LED_SEQ2_MODE   equ     B'10001010'     ;LED Sequence 2: 3x(.2s), 1s Off
LED_SEQ3_MODE   equ     B'10001011'     ;LED Sequence 3: 5x(.2s), 1s Off
LED_SLOW_MODE   equ     B'10011100'     ;LED Slow Pulsing - .3 Hz
LED_MEDIUM_MODE equ     B'10011101'     ;LED Medium Pulsing - 1 Hz
LED_FAST_MODE   equ     B'10011110'     ;LED Fast Pulsing - 3 Hz
LED_ON_MODE     equ     B'10001111'     ;LED ON Continuously


; Clear Registers 7-1Fh 
Clear_Regs   
        GOTO    Do_Clear_Regs   ;Save space in first 256 bytes

; Determine the Highest Error Level & Start Task #7 outputing the new Level
D_H_E_L   
        GOTO    Do_D_H_E_L      ;Save space in first 256 bytes

;******         ;Task #1 - Asynchronous 9600 Baud Serial Output (LOW=0)
Do_OState   
        MOVF    OState          ;if OState == 0   
        BTFSC   STATUS,Z        ;        
        GOTO    OStateS         ;           then goto Output-Start-Bit
        MOVF    RTCC,W          ;Get current time  
        MOVWF   temp            ;                 & store in tempary variable
        MOVF    First_RTCC_O,W  ;Get elapsed time; Time Unit = 2 uS 
        SUBWF   temp            ;Delta of Current Time & Orginal Time
        MOVLW   FUDGE_TIME      ;Take in account processing time to do it
        SUBWF   temp,W          ;Time within fudge factor ? 
        BTFSS   STATUS,C
        GOTO    _0005           ;Not time yet to change States so return
_0003   MOVLW   OUT_BIT_TIME    ;Past time for next out-bit ? 
        SUBWF   temp,W          
        BTFSC   STATUS,C        ;Do some delaying until it is time
        GOTO    _0004           ;It is now time to out put a bit
        MOVLW   H'04'           ;Account for loop delay 
        ADDWF   temp
        NOP                     ;   make loop delay even
        GOTO    _0003           ;Wait for exact time to output bit
_0004   MOVF    OState,W        ;Get (0-A) mode #
        ANDLW   H'0F'           ;Get only mode #  
        ADDWF   PC              ;jump to subroutine
        GOTO    OStateS         ;Serial Start Bit
        GOTO    OState0_7       ;Bit 0 
        GOTO    OState0_7       ;Bit 1 
        GOTO    OState0_7       ;Bit 2 
        GOTO    OState0_7       ;Bit 3 
        GOTO    OState0_7       ;Bit 4 
        GOTO    OState0_7       ;Bit 5 
        GOTO    OState0_7       ;Bit 6 
        GOTO    OState0_7       ;Bit 7 
        GOTO    OStateE         ;Serial Stop Bit
        GOTO    OStateL         ;Last State 
_0005   RETLW   H'00'

OStateS   
        BSF     Serial_Out      ;Serial Start Bit 
        MOVF    RTCC,W          ;Store starting time 
        MOVWF   First_RTCC_O
        MOVLW   H'0D'           ;Fudge again 
        SUBWF   First_RTCC_O
        INCF    OState          ;increment to next state 
        RETLW   H'00'

OState0_7                       ;Bit 0 - 7   
        RRF     xmt_byte        ;Move bit into C from right most bit  
        BTFSS   STATUS,C        ;        
        BCF     Serial_Out      ;
        BTFSC   STATUS,C        ;
        BSF     Serial_Out      ;
        GOTO    OS_End
OStateE   
        BCF     Serial_Out      ;Serial Stop Bit 
OS_End  MOVLW   OUT_BIT_TIME    ;Adjust out the cumulation of error   
        ADDWF   First_RTCC_O
        INCF    OState          ;increment to next state 
        RETLW   H'00'
OStateL   
        CLRF    OState          ;Ready to send next byte out  
        BCF     OState_B        ;Serial Out not active  
        RETLW   H'00'

;******         ;Task #7 - Output Highest Level Indication on LED 
Do_LED   
        BTFSC   LED_NEW_B       ;Initialize regs if change in modes 
        GOTO    LED_NEW
        INCF    LED_C           ;Inc Counter - Time Unit = 131072 uS 
        MOVF    LED_Mode,W      ;Get (0-7) mode #   
        ANDLW   H'07'           ;Get only mode #
        ADDWF   PC              ;jump to subroutine   
        GOTO    LED_OFF         ;LED OFF 
        GOTO    LED_SEQ1        ;LED Seq 1: 1 short pulse & pause  
        GOTO    LED_SEQ2        ;LED Seq 2: 2 short pulses & pause
        GOTO    LED_SEQ3        ;LED Seq 3: 3 short pulses & pause 
        GOTO    LED_SLOW        ;LED Slow Pulsing - .3 Hz 
        GOTO    LED_MEDIUM      ;LED Medium Pulsing - 1 Hz
        GOTO    LED_FAST        ;LED Fast Pulsing - 3 Hz 
        GOTO    LED_ON          ;LED ON Continuously 
_0012   RETLW   H'00'
;------
LED_OFF   
        BCF     LED             ;Turn off LED 
        BCF     LED_B           ;LED must be off 
        CLRF    LED_C           ;Reset Counter - LED_C = 0 
        RETLW   H'00'
;------
LED_NEW   
        BCF     LED_NEW_B       ;Done initializing
LED_ON   
        BSF     LED             ;Turn on LED
        CLRF    LED_C           ;Reset Counter - LED_C = 0    
        RETLW   H'00'
;------
LED_SLOW   
        MOVLW   H'0C'           ;.3Hz @ 50% Duty  
        MOVWF   temp
LED_S   MOVF    temp,W          ;Check LED_C if time, .3Hz @ 50% Duty     
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        MOVLW   H'10'
        XORWF   LED_Mode        ;Switch states 
        BTFSS   LED_Mode,4      ;Now make LED same state
        BCF     LED
        BTFSC   LED_Mode,4
        BSF     LED
        CLRF    LED_C           ;Reset LED_C
        RETLW   H'00'
;------
LED_MEDIUM   
        MOVLW   H'04'           ;1Hz @ 50% Duty 
        MOVWF   temp             
        GOTO    LED_S           ;Go do it
;------
LED_FAST   
        MOVLW   H'01'           ;3Hz @ 50% Duty 
        MOVWF   temp
        GOTO    LED_S           ;Go do it 
;------
LED_SEQ1                        ;.2 ON, 1 OFF  
        BTFSS   LED_Mode,4      ;Skip if bit is high 
        GOTO    ON1             ;Go do it 
        GOTO    OFF3            ;Go do it 
;------
LED_SEQ2                        ;.2 ON, .2 OFF, .2 ON, 1 OFF 
        MOVF    LED_Mode,W
        MOVWF   temp
        MOVLW   H'30'           ;Get sequence # only  
        ANDWF   temp
        SWAPF   temp            ;swap nibbles 
        MOVF    temp,W          ;get nibble for offset 
        ADDWF   PC              ;Table jump calculation  
        GOTO    ON1             ;LED is on, check if time to change 
        GOTO    OFF2            ;LED is off, check if time to change
        GOTO    ON1             ;LED is on, check if time to change 
        GOTO    OFF3            ;LED is off, check if time to change
;------
LED_Exit   
        MOVLW   H'10'           ;Inc Seq #   
        ADDWF   LED_Mode
        CLRF    LED_C           ;Reset LED_C 
        RETLW   H'00'
ON1     
        MOVLW   H'02'           ;Check LED_C if time, .2 sec-on  
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BCF     LED             ;Turn off LED  
        GOTO    LED_Exit
OFF2   
        MOVLW   H'02'           ;Check LED_C if time, .2 sec-on
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BSF     LED             ;Turn on LED 
        GOTO    LED_Exit
OFF3   
        MOVLW   H'08'           ;Check LED_C if time, 1 sec-off 
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BSF     LED             ;Turn on LED 
        MOVLW   H'F0'
        IORWF   LED_Mode        ;Cause (Seq# & NEW) to overflow to 0
        GOTO    LED_Exit
LED_SEQ3                        ;.2 ON, .2 OFF, .2 ON, .2 OFF, .2 ON, 1 OFF   
        MOVF    LED_Mode,W      ;Get LED info 
        MOVWF   temp
        MOVLW   H'70'           ;Get sequence # only 
        ANDWF   temp            
        SWAPF   temp            ;swap nibbles
        MOVF    temp,W          ;get nibble for offset 
        ADDWF   PC              ;Table jump calculation 
        GOTO    ON1             ;LED is on check if time to change 
        GOTO    OFF2            ;LED is off check if time to change 
        GOTO    ON1             ;LED is on check if time to change
        GOTO    OFF2            ;LED is off check if time to change 
        GOTO    ON1             ;LED is on check if time to change
        GOTO    OFF3            ;LED is off check if time to change 

;****           Quick Check of Tasks #1, #2 and #3
QCheck_T123                     
        ;Task #1 - Asynchronous 9600 Baud Serial Output (LOW=0)
        BTFSS   OState_B        ;if not outputing now then skip call
        GOTO    T2
        CALL    Do_OState       ;Go Do Task #1
        
        ;Task #2 - Asynchronous 4800 Baud Serial Input (LOW=0) 
T2      BTFSC   IState1_B       ;if already started then call
        GOTO    _0029
        BTFSC   Serial_IN_1     ;if Start bit ? then call 
        GOTO    _0029
        GOTO    T3
_0029   CALL    Do_I1State      ;Go Do Task #2
                
        ;Task #3 - Asynchronous 4800 Baud Serial Input (LOW=0)
T3      BTFSC   IState2_B       ;if already started then call
        GOTO    _0031
        BTFSC   Serial_IN_2     ;if Start bit ? then call 
        GOTO    _0031
        RETLW   H'00'
_0031   CALL    Do_I2State      ;Go Do Task #3
        RETLW   H'00'

;******         ;Task #2 - Asynchronous 4800 Baud Serial Input (LOW=0)
Do_I1State   
        MOVF    IState1         ;if IState1 == 0  
        BTFSC   STATUS,Z        ;               then Do Start Bit 
        GOTO    I1StateS
        MOVF    RTCC,W          ;Get current time
        MOVWF   temp
        MOVF    First_RTCC_I1,W ;Get elapsed time; Time Unit = 2 uS 
        SUBWF   temp
        MOVF    nbti1,W         ;Past time for next input bit ? 
        SUBWF   temp,W
        BTFSS   STATUS,C
        GOTO    _0033
        MOVF    IState1,W       ;Get (0-B) mode #
        ANDLW   H'0F'           ;Get only mode #  
        ADDWF   PC              ;jump to subroutine
        GOTO    I1StateS        ;Serial Start Bit 
        GOTO    I1State2        ;1/2 of Start Bit - see if False Start
        GOTO    I1State0_7      ;Bit 0  
        GOTO    I1State0_7      ;Bit 1  
        GOTO    I1State0_7      ;Bit 2  
        GOTO    I1State0_7      ;Bit 3  
        GOTO    I1State0_7      ;Bit 4  
        GOTO    I1State0_7      ;Bit 5  
        GOTO    I1State0_7      ;Bit 6  
        GOTO    I1State0_7      ;Bit 7  
        GOTO    I1StateE        ;Serial Stop Bit 
        GOTO    I1StateL        ;Last State - End of Stop Bit
_0033   
        RETLW   H'00'

;******         ;Task #3 - Asynchronous 4800 Baud Serial Input (LOW=0)
Do_I2State   
        MOVF    IState2         ;if IState1 == 0
        BTFSC   STATUS,Z        ;               then Do Start Bit 
        GOTO    I2StateS
        MOVF    RTCC,W          ;Get current time
        MOVWF   temp
        MOVF    First_RTCC_I2,W ;Get elapsed time; Time Unit = 2 uS
        SUBWF   temp
        MOVF    nbti2,W         ;Past time for next input bit ? 
        SUBWF   temp,W
        BTFSS   STATUS,C
        GOTO    _0035
        MOVF    IState2,W       ;Get (0-B) mode #
        ANDLW   H'0F'           ;Get only mode #  
        ADDWF   PC              ;jump to subroutine 
        GOTO    I2StateS        ;Serial Start Bit 
        GOTO    I2StateS2       ;1/2 of Start Bit - see if False Start  
        GOTO    I2State0_7      ;Bit 0
        GOTO    I2State0_7      ;Bit 1
        GOTO    I2State0_7      ;Bit 2
        GOTO    I2State0_7      ;Bit 3
        GOTO    I2State0_7      ;Bit 4
        GOTO    I2State0_7      ;Bit 5
        GOTO    I2State0_7      ;Bit 6
        GOTO    I2State0_7      ;Bit 7
        GOTO    I2StateE        ;Serial Stop Bit 
        GOTO    I2StateL        ;Last State - End of Stop Bit 
_0035   RETLW   H'00'

;***            ;Subroutines for Task #2
I1StateS                        ;Start Bit - Setup timing variables
        BSF     IState1_B       ;Serial Input Active 
        MOVF    RTCC,W          ;Store starting time
        MOVWF   First_RTCC_I1
        MOVLW   H'0D'           ;Fudge again 
        SUBWF   First_RTCC_I1
        MOVLW   H'32'           ;Time delay = 1/2 bit time 
        MOVWF   nbti1
        INCF    IState1         ;Increment to next state  
        RETLW   H'00'
I1State2                        ;Check if still a Start Bit
        BTFSS   Serial_IN_1     ;False Start Error ? 
        GOTO    FS_Error_1
        BCF     FS_Flag_1       ;Start Bit OK  
        MOVF    nbti1,W         ;Adjust out the error
        ADDWF   First_RTCC_I1
        MOVLW   IN_BIT_TIME     ;Time Delay = full bit time
        MOVWF   nbti1
        INCF    IState1         ;increment to next state
        RETLW   H'00'
I1State0_7                      ;Bit 0 - 7 
        BTFSS   Serial_IN_1     ;Move Input bit into C    
        BCF     STATUS,C
        BTFSC   Serial_IN_1
        BSF     STATUS,C
        RRF     rcv_byte_1      ;Move C into left most bit 
        MOVF    nbti1,W
        ADDWF   First_RTCC_I1   ;Adjust out the error
        INCF    IState1         ;increment to next state 
        RETLW   H'00'
I1StateE                        ;Check if we have a proper Stop Bit
        BTFSC   Serial_IN_1     ;Frame Error
        GOTO    F_Error_1
        BCF     FE_Flag_1       ;Stop Bit OK
        CLRF    T_5_S_CO        ;Reset 5 Sec Timer - got a good byte 
        ;Process the msg Here !
        MOVF    rcv_byte_1,W    ;Make a copy of just received byte  
        MOVWF   RCV_Storage
        BTFSS   RCV_Got_One_B   ;Report Lost data 
        BCF     RCV_Overflow
        BTFSC   RCV_Got_One_B
        BSF     RCV_Overflow
        BSF     RCV_Got_One_B   ;We Now have a RB Value to go out 
I1StateL   
        CLRF    IState1         ;Ready to receive next byte
        BCF     IState1_B       ;Serial In not currently active
        RETLW   H'00'
FS_Error_1                      ;False Start - Shut Down Checking
        BCF     IState1_B       ;Serial Input NOT Active
        BSF     FS_Flag_1       ;False Start Error 
        GOTO    I1StateL        ;Start All Over 
F_Error_1                       ;Frame Error - Wait for End of Stop Bit 
        MOVF    nbti1,W         ;Adjust out the error
        ADDWF   First_RTCC_I1
        MOVLW   H'32'           ;Time Delay = 1/2 bit time
        MOVWF   nbti1
        BSF     FE_Flag_1       ;Frame Error for this Byte ?
        INCF    IState1         ;Increment to next state
        RETLW   H'00'

;***            ;Subroutines for Task #3
I2StateS                        ;Start Bit - Setup timing variables
        BSF     IState2_B       ;Serial Input Active  
        MOVF    RTCC,W          ;Store starting time 
        MOVWF   First_RTCC_I2
        MOVLW   H'0D'           ;Fudge again   
        SUBWF   First_RTCC_I2
        MOVLW   H'32'           ;Time delay = 1/2 bit time
        MOVWF   nbti2
        INCF    IState2         ;Increment to next state
        RETLW   H'00'
I2StateS2                       ;Check if still a Start Bit
        BTFSS   Serial_IN_2     ;False Start Error ?  
        GOTO    FS_Error_2
        BCF     FS_Flag_2       ;Start Bit OK 
        MOVF    nbti2,W         ;Adjust out the error 
        ADDWF   First_RTCC_I2
        MOVLW   IN_BIT_TIME     ;Time Delay = full bit time 
        MOVWF   nbti2
        INCF    IState2         ;increment to next state
        RETLW   H'00'
I2State0_7                      ;Bit 0 - 7 
        BTFSS   Serial_IN_2     ;Move Input bit into C
        BCF     STATUS,C
        BTFSC   Serial_IN_2
        BSF     STATUS,C
        RRF     rcv_byte_2      ;Move C into left most bit 
        MOVF    nbti2,W
        ADDWF   First_RTCC_I2   ;Adjust out the error    
        INCF    IState2         ;increment to next state 
        RETLW   H'00'
I2StateE                        ;Check if we have a proper Stop Bit
        BTFSC   Serial_IN_2     ;Frame Error
        GOTO    F_Error_2
        BCF     FE_Flag_2       ;Stop Bit OK 
        CLRF    T_5_S_CO        ;Reset 5 Sec Timer - got a good byte
        ;Process the msg Here !
        MOVF    rcv_byte_2,W    ;Make a copy of just received byte 
        MOVWF   RCV_Storage
        BTFSS   RCV_Got_One_B   ;Report Lost data 
        BCF     RCV_Overflow
        BTFSC   RCV_Got_One_B
        BSF     RCV_Overflow
        BSF     RCV_Got_One_B   ;We Now have a RB Value to go out   
I2StateL   
        CLRF    IState2         ;Ready to receive next byte 
        BCF     IState2_B       ;Serial In not currently active 
        RETLW   H'00'
FS_Error_2   
        BCF     IState2_B       ;False Start - Shut Down Checking
        BSF     FS_Flag_2       ;False Start Error 
        GOTO    I2StateL        ;Start All Over
F_Error_2                       ;Frame Error - Wait for End of Stop Bit
        MOVF    nbti2,W         ;Adjust out the error 
        ADDWF   First_RTCC_I2
        MOVLW   H'32'           ;Time Delay = 1/2 bit time 
        MOVWF   nbti2
        BSF     FE_Flag_2       ;Frame Error for this Byte ? 
        INCF    IState2         ;Increment to next state 
        RETLW   H'00'

;******         ;Code Starting point
Main   
        MOVLW   H'00'           ;What is High/Low for RA at INIT State 
        MOVWF   Port_A
        MOVLW   H'00'           ;What is High/Low for RB at INIT State 
        MOVWF   Port_B
        MOVLW   H'F9'           ;RA TRIS at INIT State  
        TRIS    5
        MOVLW   H'FF'           ;RB TRIS at INIT State 
        TRIS    6
        MOVLW   H'00'           ;RTCC/2    
        OPTION
        CALL    Clear_Regs      ;Clear Registers 7-1F - Same Memory Page
        CLRF    RTCC            ;Start timers

;Initialize Tasks
                                ;Task #1 waits for byte to output
                                ;Task #2 waits for Serial IN Start Bit
                                ;Task #3 waits for Serial IN Start Bit
                                ;Task #4 runs when Task 1 is Not
                                ;Task #5 is always running
        MOVF    Port_B,W        ;Task #6 is Initialized here 
        MOVWF   Old_RB
        MOVF    Old_RB,W        ;Make all the same initial value 
        MOVWF   Last_RB
        BSF     RB_NEW_B        ;Tell Task #4: RB byte ready to output
        MOVLW   LED_OFF_MODE
        MOVWF   LED_Mode        ;Task #7 is Started 
        BSF     T_5_S_B         ;Task #8 is Started here  
        BSF     T_5_M_B         ;Task #9 is Started here  

; Handle Task & Timer activities - Main Loop
Task_1          ;Task #1 - Asynchronous 9600 Baud Serial Output (LOW=0)
        BTFSS   OState_B        ;if not outputing now then skip call 
        GOTO    Task_2
        CALL    Do_OState       ;Go Do Task #1

Task_2          ;Task #2 - Asynchronous 4800 Baud Serial Input (LOW=0)
        BTFSC   IState1_B       ;if already started then call
        GOTO    _0053
        BTFSC   Serial_IN_1     ;if Start bit ? then call
        GOTO    _0053
        GOTO    Task_3
_0053   CALL    Do_I1State      ;Go Do Task #2

Task_3          ;Task #3 - Asynchronous 4800 Baud Serial Input (LOW=0)
        BTFSC   IState2_B       ;if already started then call 
        GOTO    _0055
        BTFSC   Serial_IN_2     ;if Start bit ? then call 
        GOTO    _0055
        GOTO    Task_4
_0055   CALL    Do_I2State      ;Go Do Task #3

Task_4          ;Task #4 - Finds next Buffered Byte to Send Out through Task 1
        BTFSC   OState_B        ;if outputing now then skip call
        GOTO    _0059
        BTFSS   RCV_Got_One_B   ;Got a NEW Received byte to send
        GOTO    _0057
        MOVF    RCV_Storage,W   ;Send just received byte 
        MOVWF   xmt_byte
        BCF     RCV_Got_One_B   ;Clear need to send old byte 
        BSF     OState_B        ;Start Task #1 & Lock Out Others 
        GOTO    _0059
_0057   BTFSS   RB_NEW_B        ;Indicates a change in RB input 
        GOTO    _0058
        MOVF    Old_RB,W        ;Send New RB value  
        MOVWF   xmt_byte
        BCF     RB_NEW_B        ;Clear need to send out newest value 
        BSF     OState_B        ;Start Task #1 & Lock Out Others
        GOTO    _0059
_0058   BTFSS   S_5_S_B         ;Serial In 5 secs of inactivity
        GOTO    _0059
        MOVLW   H'FF'           ;Tell of inactivity of Serial In 
        MOVWF   xmt_byte
        BCF     S_5_S_B         ;Clear need to send msg  
        BSF     OState_B        ;Start Task #1 & Lock Out Others
                
                ;Heart Beat - Time unit = 512 uS for Tasks #5 & #6
_0059   MOVF    RTCC,W          ;Step-up time units * 512  
        MOVWF   temp
        MOVF    Last_RTCC,W     ;Test to see if it overflowed 
        SUBWF   temp,W
        BTFSS   STATUS,C
        GOTO    Inc_Time
        MOVF    temp,W          ;unit error = < |+-512 uS| 
        MOVWF   Last_RTCC
        GOTO    Task_1
Inc_Time   
        MOVF    temp,W          ;Save current RTCC into Last_RTCC
        MOVWF   Last_RTCC

Task_5          ;Task #5 - Monitor Level Reset Input Line - Always Running ! 
        BTFSC   Level_Reset
        GOTO    Task_6
        MOVLW   LED_OFF_MODE    ;Lowest Level Indicator output
        MOVWF   LED_Mode

Task_6          ;Task #6 - Debounce 8 bit Input Sensors - Runs every 20 mS 
        INCF    T_20_mS_CO      ;Inc Counter - Time Unit = 512 uS
        MOVLW   H'27'           ;Used to debounce the input 
        SUBWF   T_20_mS_CO,W
        BTFSS   STATUS,Z
        GOTO    _0065
        CLRF    T_20_mS_CO      ;Reset T_20_mS_CO to start over again
        
        CALL    QCheck_T123     ;Quick Check of Tasks #1, #2 and #3 
        
        MOVF    Port_B,W        ;Last copy of RB same as Current ? 
        SUBWF   Last_RB,W
        BTFSC   STATUS,Z        
        GOTO    _0062
        MOVF    Port_B,W        ;  Store Current RB - diff from Last
        MOVWF   Last_RB
        GOTO    _0063
_0062   MOVF    Last_RB,W       ;New Old RB <- same value over 20 mS 
        MOVWF   Old_RB
_0063   MOVF    Old_RB          ;See if RB is now 0 
        BTFSC   STATUS,Z        ;RB == 0 ? then keep timer running 
        GOTO    _0064
        CLRF    T_5_M_LO        ;Reset 5 Min Timer 
        CLRF    T_5_M_HI        ;  still not zero yet 
_0064   CALL    D_H_E_L         ;Determine the Highest Error Level
        BTFSS   RB_NEW_B        ;Check for Lost Data Error 
        BCF     RB_Overflow
        BTFSC   RB_NEW_B
        BSF     RB_Overflow
        BSF     RB_NEW_B        ;Every 20 mS send Old_RB out
                
                ;Heart Beat - Time unit = 131072 uS for Tasks #7, #8 & #9
_0065   MOVLW   H'F9'           ;RA TRIS - refresh 
        TRIS    5
        MOVLW   H'FF'           ;RB TRIS - refresh 
        TRIS    6
        DECFSZ  cc              ;Step-up time units * 256 
        GOTO    Task_1

Task_7          ;Task 7 - Output Highest Level Indication on LED
        BTFSS   LED_B           ;Is LED active ? 
        GOTO    Task_8

        CALL    QCheck_T123     ;Quick Check of Tasks #1, #2 and #3 
        
        CALL    Do_LED          ;Handle LED timing
                                  
Task_8          ;Task #8 - 5 Second Serial Input Lack of Activity Timer
        BTFSS   T_5_S_B         ;5 Sec Timer Active ?  
        GOTO    Task_9
        INCF    T_5_S_CO        ;Inc Counter - Time Unit = 131072 uS 
        MOVLW   H'26'           ;Check T_5_S_CO if time  
        SUBWF   T_5_S_CO,W
        BTFSS   STATUS,Z
        GOTO    Task_9
        CLRF    T_5_S_CO        ;Reset T_5_S_CO 
        MOVLW   LED_ON_MODE     ;Highest Level Indicator output 
        MOVWF   LED_Mode
        BTFSS   S_5_S_B         ;Check if Lost Data Error 
        BCF     S_5_S_Overflow
        BTFSC   S_5_S_B
        BSF     S_5_S_Overflow
        BSF     S_5_S_B          ;Send notice of 5 seconds of inaction  

Task_9          ;Task #9 - 5 Min. Lack of Severe Error from Sensors Reset Timer 
        BTFSS   T_5_M_B         ;5 Min Timer Active ? 
        GOTO    Task_A
        INCF    T_5_M_LO        ;Inc LO Counter; Time Unit = 131072 uS 
        BTFSC   STATUS,Z        ;See if carry needs to be passed on ?
        INCF    T_5_M_HI        ;Inc HI Counter; Time Unit = 131072 uS
        MOVLW   H'08'           ;#2288<  Check T_5_M_HI if time  
        SUBWF   T_5_M_HI,W
        BTFSS   STATUS,Z
        GOTO    Task_A
        MOVLW   H'F0'           ;#2288>  Check T_5_M_LO if time 
        SUBWF   T_5_M_LO,W
        BTFSS   STATUS,Z
        GOTO    Task_A
        CLRF    T_5_M_LO        ;Reset T_5_M_LO 
        CLRF    T_5_M_HI        ;Reset T_5_M_HI 
        MOVLW   LED_OFF_MODE    ;Lowest Level Indicator output 
        MOVWF   LED_Mode
Task_A   
        GOTO    Task_1          ;Loop Forever

;****
Do_D_H_E_L      ; Determine the Highest Error Level & Start Task #7
        MOVLW   H'07'           ;Check top 7 bits 
        MOVWF   temp
        MOVF    Old_RB,W        ;Get copy of 7 debounced Sensor Input
        MOVWF   Last_RB
_0070   RLF     Last_RB         ;Put top bit into C bit
        BTFSC   STATUS,C        ;Check if C bit is set 
        GOTO    _0072
        DECFSZ  temp            ;Continue to check lesser bits
        GOTO    _0070
_0071   MOVF    Port_B,W        ;Restore current value of RB 
        MOVWF   Last_RB
        RETLW   H'00'
_0072   MOVF    LED_Mode,W      ;Get current Level Indicator 
        ANDLW   H'07'           ;Get only      "       "  
        MOVWF   Last_RB         ;Store into a tempary register 
        MOVF    temp,W          ;Check if already at this Level
        SUBWF   Last_RB,W
        BTFSC   STATUS,C
        GOTO    _0071
        MOVLW   H'88'           ;Start to build LED_Mode 
        IORWF   temp,W          ;Put new Level Indicator into reg 
        MOVWF   LED_Mode        ;Store new LED Mode 
        GOTO    _0071

Do_Clear_Regs   ; Clear Registers 7-1Fh
        MOVLW   H'1F'           ;First regs to clear
        MOVWF   FSR
Loop_C  CLRF    INDIR           ;Clear reg 
        DECF    FSR             ;point to next reg to clear 
        MOVLW   H'E7'           ;Dec temp, jump if not done
        SUBWF   FSR,W
        BTFSC   STATUS,C
        GOTO    Loop_C
        CLRF    FSR             ;Lastly clear FSR reg 
        RETLW   H'00'

        ORG     H'1FF'          ;RESET to Main
        GOTO    Main

        END

