; 'Remote Alarm64' V1.00 Designed by Myriad Development Co. - Jerry Farmer
;       PIC16C64, 4MHz Crystal, WatchDog Timer OFF, MPASM instruction set

        list    p=16C64,t=ON,c=132

        include "PIC16C64.H"

; B Register Definitions
Serial_IN_1     equ     0               ;Serial Input #1 - 8 bits - INT pin
                                        ;RB.7 - RB.1 == Input from Sensors
RB_TRIS         equ     B'11111111'     ;RB TRIS at INIT State == all input
RB_MASK         equ     B'00000000'     ;What is High/Low for RB at INIT State

; A Register Definitions - Programmable Inputs
Level_Reset     equ     0               ;PORTA.0 - Reset Level Indicator
LED             equ     1               ;LED Output - Level/State Indicator
Serial_Out      equ     2               ;Serial Output - 8 bits + passwords
PWM_Out         equ     3               ;PWM Output - 8 bits ON, 8 bits OFF

RA_TRIS         equ     B'11110001'     ;RA TRIS at INIT State
RA_MASK         equ     B'00000000'     ;What is High/Low for RA at INIT State

; Register Files
temp            equ     20h             ;Tempary holding register - PIC16C54/56
tmp             equ     21h             ;Tempary reg
Temp_W          equ     22h             ;Interrupt storage of W
Temp_Stat       equ     23h             ;Interrupt storage of STATUS
Temp_FSR        equ     24h             ;Interrupt storage of FSR
T_B             equ     25h             ;Indicates which Timer(s) are Active = 1
FLAGS           equ     26h             ;Error Flags
LED_Mode        equ     27h             ;(0-2)=Mode, 3=LED_B, (4-6)=Seq #, 7=NEW
OState          equ     28h             ;Serial Out State
IState1         equ     29h             ;Serial In #1 State
cc              equ     2Ah             ;256 * TMR0 time
T_5_M_LO        equ     2Bh             ;5 Min Timer Counter - Low
T_5_M_HI        equ     2Ch             ;5 Min Timer Counter - High
T_5_S_CO        equ     2Dh             ;5 Second Timer - lack of Serial Input
T_20_mS_CO      equ     2Eh             ;20 mS Timer - used for debouncing
T_PWM_CO        equ     2Fh             ;PWM Counter
LED_C           equ     30h             ;LED Counter
xmt_byte        equ     31h             ;Serial xmit byte - destroyed in use
rcv_byte_1      equ     32h             ;Receive Serial #1 In byte
RCV_Storage     equ     33h             ;Long term storage of rcv_byte #1
Old_RB          equ     34h             ;Oldest/Master copy of RB
Last_RB         equ     35h             ;Last copy of RB
PWM_ON          equ     36h             ;PWM ON Counter
PWM_OFF         equ     37h             ;PWM OFF Counter
PWM_tmp         equ     38h             ;PWM tempary counter

; Indicates which Timer(s) are Active = 1 & Flags - T_B
OState_B        equ     0               ;Serial Out Active Bit
IState1_B       equ     1               ;Serial IN #1 Active Bit
T_5_S_B         equ     2               ;5 Second Timer Active Bit
T_5_M_B         equ     3               ;5 Min Timer Active Bit
RCV_Got_One_B   equ     4               ;Got a NEW Received byte to send out
RB_NEW_B        equ     5               ;Indicates a change in RB input
S_5_S_B         equ     6               ;Serial In 5 secs of inactivity
T_PWM_B         equ     7               ;PWM Activity Bit

; Error Flags - FLAGS
FS_Flag_1       equ     0               ;Serial #1 IN had a False Start Error
FE_Flag_1       equ     1               ;Last Serial #1 IN had a Frame Error
RCV_Overflow    equ     2               ;Lost Serial Input Byte - too Slow
RB_Overflow     equ     3               ;Lost RB Input Byte - too Slow
S_5_S_Overflow  equ     4               ;Lost '5S Inactivity' msg - too Slow
Time_Bit        equ     5               ;Indicate 512 uS has passed

;Equates for LED Task #7 - LED_Mode
LED_B           equ     3               ;LED is active - LED_Mode.3
LED_NEW_B       equ     7               ;LED has just changed Modes = 1
LED_OFF_MODE    equ     B'00001000'     ;LED OFF
LED_SEQ1_MODE   equ     B'10001001'     ;LED Sequence 1: .2s On, 1s Off
LED_SEQ2_MODE   equ     B'10001010'     ;LED Sequence 2: 3x(.2s), 1s Off
LED_SEQ3_MODE   equ     B'10001011'     ;LED Sequence 3: 5x(.2s), 1s Off
LED_SLOW_MODE   equ     B'10011100'     ;LED Slow Pulsing - .3 Hz
LED_MEDIUM_MODE equ     B'10011101'     ;LED Medium Pulsing - 1 Hz
LED_FAST_MODE   equ     B'10011110'     ;LED Fast Pulsing - 3 Hz
LED_ON_MODE     equ     B'10001111'     ;LED ON Continuously

        ORG     0               ;Reset Vector

        GOTO    Main

        ORG     4               ;Interrupt Vector

        GOTO    Interrupt

;******         Task #1 - Asynchronous 9600 Baud Serial Output (LOW=0)
Do_OState   
        MOVF    OState,W        ;Get (0-A) mode #
        ANDLW   H'0F'           ;Get only mode #  
        ADDWF   PCL             ;jump to subroutine
        GOTO    OStateS         ;Serial Start Bit
        GOTO    OState0_7       ;Bit 0 
        GOTO    OState0_7       ;Bit 1 
        GOTO    OState0_7       ;Bit 2 
        GOTO    OState0_7       ;Bit 3 
        GOTO    OState0_7       ;Bit 4 
        GOTO    OState0_7       ;Bit 5 
        GOTO    OState0_7       ;Bit 6 
        GOTO    OState0_7       ;Bit 7 
        GOTO    OStateE         ;Serial Stop Bit
        GOTO    OStateL         ;Last State 
        RETURN

OStateS   
        MOVLW   H'00'           ;Post & Pre 1=1 & OFF
        MOVWF   T2CON
        BSF     STATUS,RP0      ;Point to BANK 1
        MOVLW   H'68' - H'15'   ;104uS - 9600 Baud & adjust to Latency  
        MOVWF   PR2
        BCF     STATUS,RP0      ;Point to BANK 0
        CLRF    TMR2            ;Init to 0
        BCF     PIR1,TMR2IF     ;Clear Timer 2 Flag so as to start fresh
        BSF     PORTA,Serial_Out  ;Output 'Serial Start Bit' starting Now
        BSF     T2CON,TMR2ON    ;Start Timer 2
        INCF    OState          ;inc to next state BEFORE allowing interrupts 
        
        BSF     STATUS,RP0      ;Point to BANK 1
        BSF     PIE1,TMR2IE     ;Allow for Timer 2 interrupts
        BCF     STATUS,RP0      ;Point to BANK 0
        RETURN

OState0_7                       ;Bit 0 - 7   
        RRF     xmt_byte        ;Move bit into C from right most bit  
        BTFSS   STATUS,C          ;Output C bit        
        BCF     PORTA,Serial_Out  ;
        BTFSC   STATUS,C          ;
        BSF     PORTA,Serial_Out  ;
        GOTO    OS_End
OStateE   
        BCF     PORTA,Serial_Out  ;Serial Stop Bit 
OS_End  BCF     PIR1,TMR2IF     ;Clear Timer 2 Flag so as to start fresh
        BSF     STATUS,RP0      ;Point to BANK 1
        MOVLW   H'68' + H'4'    ;104uS - 9600 Baud & adjust to Latency
        MOVWF   PR2
        BCF     STATUS,RP0      ;Point to BANK 0
        INCF    OState          ;increment to next state 
        RETURN
OStateL   
        BSF     STATUS,RP0      ;Point to BANK 1
        BCF     PIE1,TMR2IE     ;Do NOT Allow Timer 2 interrupts
        BCF     STATUS,RP0      ;Point to BANK 0
        
        BCF     PIR1,TMR2IF     ;Clear Timer 2 Flag so as to start fresh
        BCF     T2CON,TMR2ON    ;Stop Timer 2
        CLRF    OState          ;Ready to send next byte out  
        BCF     T_B,OState_B    ;Serial Out not active  
        RETURN

;******         Task #7 - Output Highest Level Indication on LED 
Do_LED   
        BTFSC   LED_Mode,LED_NEW_B  ;Initialize regs if change in modes 
        GOTO    LED_NEW
        INCF    LED_C           ;Inc Counter - Time Unit = 131072 uS 
        MOVF    LED_Mode,W      ;Get (0-7) mode #   
        ANDLW   H'07'           ;Get only mode #
        ADDWF   PCL             ;jump to subroutine   
        GOTO    LED_OFF         ;LED OFF 
        GOTO    LED_SEQ1        ;LED Seq 1: 1 short pulse & pause  
        GOTO    LED_SEQ2        ;LED Seq 2: 2 short pulses & pause
        GOTO    LED_SEQ3        ;LED Seq 3: 3 short pulses & pause 
        GOTO    LED_SLOW        ;LED Slow Pulsing - .3 Hz 
        GOTO    LED_MEDIUM      ;LED Medium Pulsing - 1 Hz
        GOTO    LED_FAST        ;LED Fast Pulsing - 3 Hz 
        GOTO    LED_ON          ;LED ON Continuously 
_0012   RETURN
;------
LED_OFF   
        BCF     PORTA,LED       ;Turn off LED 
        BCF     LED_Mode,LED_B  ;LED must be off 
        CLRF    LED_C           ;Reset Counter - LED_C = 0 
        RETURN
;------
LED_NEW   
        BCF     LED_Mode,LED_NEW_B  ;Done initializing
LED_ON   
        BSF     PORTA,LED       ;Turn on LED
        CLRF    LED_C           ;Reset Counter - LED_C = 0    
        RETURN
;------
LED_SLOW   
        MOVLW   H'0C'           ;.3Hz @ 50% Duty  
        MOVWF   temp
LED_S   MOVF    temp,W          ;Check LED_C if time, .3Hz @ 50% Duty     
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        MOVLW   H'10'
        XORWF   LED_Mode        ;Switch states 
        BTFSS   LED_Mode,4      ;Now make LED same state
        BCF     PORTA,LED
        BTFSC   LED_Mode,4
        BSF     PORTA,LED
        CLRF    LED_C           ;Reset LED_C
        RETURN
;------
LED_MEDIUM   
        MOVLW   H'04'           ;1Hz @ 50% Duty 
        MOVWF   temp             
        GOTO    LED_S           ;Go do it
;------
LED_FAST   
        MOVLW   H'01'           ;3Hz @ 50% Duty 
        MOVWF   temp
        GOTO    LED_S           ;Go do it 
;------
LED_SEQ1                        ;.2 ON, 1 OFF  
        BTFSS   LED_Mode,4      ;Skip if bit is high 
        GOTO    ON1             ;Go do it 
        GOTO    OFF3            ;Go do it 
;------
LED_SEQ2                        ;.2 ON, .2 OFF, .2 ON, 1 OFF 
        MOVF    LED_Mode,W
        MOVWF   temp
        MOVLW   H'30'           ;Get sequence # only  
        ANDWF   temp
        SWAPF   temp            ;swap nibbles 
        MOVF    temp,W          ;get nibble for offset 
        ADDWF   PCL             ;Table jump calculation  
        GOTO    ON1             ;LED is on, check if time to change 
        GOTO    OFF2            ;LED is off, check if time to change
        GOTO    ON1             ;LED is on, check if time to change 
        GOTO    OFF3            ;LED is off, check if time to change
;------
LED_Exit   
        MOVLW   H'10'           ;Inc Seq #   
        ADDWF   LED_Mode
        CLRF    LED_C           ;Reset LED_C 
        RETURN
ON1     
        MOVLW   H'02'           ;Check LED_C if time, .2 sec-on  
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BCF     PORTA,LED       ;Turn off LED  
        GOTO    LED_Exit
OFF2   
        MOVLW   H'02'           ;Check LED_C if time, .2 sec-on
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BSF     PORTA,LED       ;Turn on LED 
        GOTO    LED_Exit
OFF3   
        MOVLW   H'08'           ;Check LED_C if time, 1 sec-off 
        SUBWF   LED_C,W
        BTFSS   STATUS,Z
        GOTO    _0012
        BSF     PORTA,LED       ;Turn on LED 
        MOVLW   H'F0'
        IORWF   LED_Mode        ;Cause (Seq# & NEW) to overflow to 0
        GOTO    LED_Exit
LED_SEQ3                        ;.2 ON, .2 OFF, .2 ON, .2 OFF, .2 ON, 1 OFF   
        MOVF    LED_Mode,W      ;Get LED info 
        MOVWF   temp
        MOVLW   H'70'           ;Get sequence # only 
        ANDWF   temp            
        SWAPF   temp            ;swap nibbles
        MOVF    temp,W          ;get nibble for offset 
        ADDWF   PCL             ;Table jump calculation 
        GOTO    ON1             ;LED is on check if time to change 
        GOTO    OFF2            ;LED is off check if time to change 
        GOTO    ON1             ;LED is on check if time to change
        GOTO    OFF2            ;LED is off check if time to change 
        GOTO    ON1             ;LED is on check if time to change
        GOTO    OFF3            ;LED is off check if time to change 

;******         Task #2 - Asynchronous 4800 Baud Serial Input (LOW=0)
Do_I1State   
        MOVF    IState1,W       ;Get (0-B) mode #
        ANDLW   H'0F'           ;Get only mode #  
        ADDWF   PCL             ;jump to subroutine
        GOTO    I1StateS        ;Serial Start Bit 
        GOTO    I1State2        ;1/2 of Start Bit - see if False Start
        GOTO    I1State0_7      ;Bit 0  
        GOTO    I1State0_7      ;Bit 1  
        GOTO    I1State0_7      ;Bit 2  
        GOTO    I1State0_7      ;Bit 3  
        GOTO    I1State0_7      ;Bit 4  
        GOTO    I1State0_7      ;Bit 5  
        GOTO    I1State0_7      ;Bit 6  
        GOTO    I1State0_7      ;Bit 7  
        GOTO    I1StateE        ;Serial Stop Bit 
        GOTO    I1StateL        ;Last State - End of Stop Bit
        RETURN

;***            Subroutines for Task #2
I1StateS                        ;Start Bit - Setup timing variables
        BCF     INTCON,INTE     ;Disable detecting changes on INT pin
        BCF     INTCON,INTF     ;Clear Interrupting Flag
        MOVLW   H'00'           ;Internal Clk, Pre 1=1 & OFF
        MOVWF   T1CON
        CLRF    TMR1L           ;Calculate (0 - #) of counts until roll-over
        MOVLW   H'68' - H'1E'   ;208us/2 = 4800 Baud & adjust to Latency
        SUBWF   TMR1L
        CLRF    TMR1H
        DECF    TMR1H           ;H'FF'
        BCF     PIR1,TMR1IF     ;Clear Timer 1 Flag so as to start fresh
        BSF     T1CON,TMR1ON    ;Start Timer 1
        INCF    IState1         ;inc to next state BEFORE allowing interrupts 
        BSF     T_B,IState1_B   ;Serial Input Active 
        
        BSF     STATUS,RP0      ;Point to BANK 1
        BSF     PIE1,TMR1IE     ;Allow for Timer 1 interrupts
        BCF     STATUS,RP0      ;Point to BANK 0
        RETURN
I1State2                        ;Check if still a Start Bit
        BTFSS   PORTB,Serial_IN_1  ;False Start Error ? 
        GOTO    FS_Error_1
        BCF     FLAGS,FS_Flag_1 ;Start Bit OK  
        BCF     T1CON,TMR1ON    ;Stop Timer 1
        CLRF    TMR1L
        MOVLW   H'D0' - H'26'   ;208us = 4800 Baud & adjust to Latency
        SUBWF   TMR1L
        CLRF    TMR1H
        DECF    TMR1H           ;H'FF'
        BCF     PIR1,TMR1IF     ;Clear Timer 1 Flag so as to start fresh
        BSF     T1CON,TMR1ON    ;Start Timer 1
        INCF    IState1         ;increment to next state
        RETURN
I1State0_7                      ;Bit 0 - 7 
        BTFSS   PORTB,Serial_IN_1  ;Move Input bit into C    
        BCF     STATUS,C
        BTFSC   PORTB,Serial_IN_1
        BSF     STATUS,C
        RRF     rcv_byte_1      ;Move C into left most bit 
        BCF     T1CON,TMR1ON    ;Stop Timer 1
        CLRF    TMR1L
        MOVLW   H'D0' - H'26'   ;208us = 4800 Baud & adjust to Latency
        SUBWF   TMR1L
        CLRF    TMR1H
        DECF    TMR1H           ;H'FF'
        BCF     PIR1,TMR1IF     ;Clear Timer 1 Flag so as to start fresh
        BSF     T1CON,TMR1ON    ;Start Timer 1
        INCF    IState1         ;increment to next state 
        RETURN
I1StateE                        ;Check if we have a proper Stop Bit
        BTFSC   PORTB,Serial_IN_1  ;Frame Error
        GOTO    F_Error_1
        BCF     FLAGS,FE_Flag_1 ;Stop Bit OK
        CLRF    T_5_S_CO        ;Reset 5 Sec Timer - got a good byte 
        ;Process the msg Here !
        MOVF    rcv_byte_1,W    ;Make a copy of just received byte  
        MOVWF   RCV_Storage
        BTFSS   T_B,RCV_Got_One_B  ;Report Lost data 
        BCF     FLAGS,RCV_Overflow
        BTFSC   T_B,RCV_Got_One_B
        BSF     FLAGS,RCV_Overflow
        BSF     T_B,RCV_Got_One_B  ;We Now have a RB Value to go out 
I1StateL   
        BCF     T1CON,TMR1ON    ;Stop Timer 1
        BSF     STATUS,RP0      ;Point to BANK 1
        BSF     PIE1,TMR1IE     ;Allow for Timer 1 interrupts
        BCF     STATUS,RP0      ;Point to BANK 0
        BCF     PIR1,TMR1IF     ;Clear Timer 1 Flag so as to start fresh
        
        CLRF    IState1         ;Ready to receive next byte
        BCF     T_B,IState1_B   ;Serial In not currently active
        
        BCF     INTCON,INTF     ;Clear Interrupting Flag
        BSF     INTCON,INTE     ;Enable detecting changes on INT pin
        RETURN
FS_Error_1                      ;False Start - Shut Down Checking
        BCF     T_B,IState1_B   ;Serial Input NOT Active
        BSF     FLAGS,FS_Flag_1  ;False Start Error 
        GOTO    I1StateL        ;Start All Over 
F_Error_1                       ;Frame Error - Wait for End of Stop Bit 
        BSF     FLAGS,FE_Flag_1 ;Frame Error for this Byte ?
        INCF    IState1         ;Increment to next state
        BCF     T1CON,TMR1ON    ;Stop Timer 1
        CLRF    TMR1L
        MOVLW   H'68' - H'1E'   ;208us/2 = 4800 Baud & adjust to Latency
        SUBWF   TMR1L
        CLRF    TMR1H
        DECF    TMR1H           ;H'FF'
        BCF     PIR1,TMR1IF     ;Clear Timer 1 Flag so as to start fresh
        BSF     T1CON,TMR1ON    ;Start Timer 1
        RETURN

;******         Code Starting point
Main   
        CLRF    STATUS
        CLRF    FSR
        CLRF    TMR0            ;Clear Timer0
        MOVLW   H'00'           ;What is High/Low for RA at RESET State 
        MOVWF   PORTA
        MOVWF   PORTB
        MOVWF   PORTC
        MOVWF   PORTD
        MOVWF   PORTE
        CLRF    PCLATH
        MOVLW   H'60'           ;/GIE,PEIE,RTIE,/INTE,/RBIE,/RTIF,/INTF,/RBIF
        MOVWF   INTCON
        CLRF    PIR1            ;Timer 2 Flag cleared
        CLRF    TMR1L
        CLRF    TMR1H
        CLRF    T1CON           ;Timer 1 OFF until ready for input 
        CLRF    TMR2
        CLRF    T2CON           ;Timer 2 OFF until have byte to output
        CLRF    SSPBUF
        CLRF    SSPCON
        CLRF    CCPR1L
        CLRF    CCPR1H
        CLRF    CCP1CON
        
        BSF     STATUS,RP0      ;Point to BANK 1
        MOVLW   H'40'           ;TMR0/2 & Interrupt on Rising edge of INT    
        MOVWF   OPTION          ;Load OPTION reg
        MOVLW   RA_TRIS
        MOVWF   TRISA
        MOVLW   H'FF'           ;RB TRIS at RESET State  
        MOVWF   TRISB
        MOVWF   TRISC
        MOVWF   TRISD
        MOVLW   H'07'           ;PSPMODE=0
        MOVWF   TRISE
        CLRF    PIE1            ;Timer 2 Interrupt disabled
        BSF     PCON,POR        
        CLRF    PR2
        CLRF    SSPADD
        CLRF    SSPSTAT
        
        BCF     STATUS,RP0      ;Point to BANK 0
        CALL    Clear_Regs      ;Clear Registers 20-7F, A0-C0 Memory Pages

;Initialize Tasks
                                ;Task #1 waits for byte to output
                                ;Task #2 waits for Serial IN Start Bit
        MOVLW   H'31'           ;Task #3 is initialized for square pulses
        MOVWF   PWM_ON          ;   "     25 mS ON
        MOVLW   H'31'           ;   "              Period = 50 mS, DS= 50%
        MOVWF   PWM_OFF         ;   "     25 mS OFF
        CLRF    T_PWM_CO        ;   "
        MOVF    PWM_ON,W        ;mov PWM_tmp,PWM_ON
        MOVWF   PWM_tmp         ; "
        BSF     PORTA,PWM_Out   ;Start Outputing ON
        BSF     T_B,T_PWM_B     ;   "
                                ;Task #4 runs when Task 1 is Not
                                ;Task #5 is always running
        MOVF    PORTB,W         ;Task #6 is Initialized here 
        MOVWF   Old_RB
        MOVF    Old_RB,W        ;Make all the same initial value 
        MOVWF   Last_RB
        BSF     T_B,RB_NEW_B    ;Tell Task #4: RB byte ready to output
        MOVLW   LED_OFF_MODE
        MOVWF   LED_Mode        ;Task #7 is Started 
        BSF     T_B,T_5_S_B     ;Task #8 is Started here  
        BSF     T_B,T_5_M_B     ;Task #9 is Started here  

        BSF     INTCON,GIE      ;Enable Global Interrupts

; Handle Task & Timer activities - Main Loop done in background
Inc_Time        ;Heart Beat - Time unit = 512 uS for Tasks #5 & #6 
        BTFSS   FLAGS,Time_Bit  ;Idle Task - wait until 512 uS has gone by
        GOTO    Inc_Time
        BCF     FLAGS,Time_Bit  ;Reset for next indicator from TMR0 Interrupt

Task_3          ;Task #3 - PWM, Period = (PWM_ON + PWM_OFF) * 512uS
        BTFSS   T_B,T_PWM_B     ;if NOT outputing now then skip call
        GOTO    Task_4
        INCF    T_PWM_CO        ;Inc count of time
        MOVF    PWM_tmp,W       ;cjne T_PWM_CO,PWM_tmp,Task_4
        SUBWF   T_PWM_CO,W      ; "
        BTFSS   STATUS,Z        ; "
        GOTO    Task_4          ; "
        CLRF    T_PWM_CO        ;Reset timer
        BTFSC   PORTA,PWM_Out
        GOTO    T3_1
        BSF     PORTA,PWM_Out   ;Change Output State
        MOVF    PWM_ON,W        ;mov PWM_tmp,PWM_ON
        MOVWF   PWM_tmp         ; "
        GOTO    Task_4
T3_1
        BCF     PORTA,PWM_Out   ;Change Output State
        MOVF    PWM_OFF,W       ;mov PWM_tmp,PWM_OFF
        MOVWF   PWM_tmp       ; "
        
Task_4          ;Task #4 - Finds next Buffered Byte to Send Out through Task 1
        BTFSC   T_B,OState_B    ;if outputing now then skip call
        GOTO    Task_5
        BTFSS   T_B,RCV_Got_One_B  ;Got a NEW Received byte to send
        GOTO    _0057
        MOVF    RCV_Storage,W   ;Send just received byte 
        MOVWF   xmt_byte
        BCF     T_B,RCV_Got_One_B  ;Clear need to send old byte 
        BSF     T_B,OState_B    ;Start Task #1 & Lock Out Others 
        GOTO    T4_S
_0057   BTFSS   T_B,RB_NEW_B    ;Indicates a change in RB input 
        GOTO    _0058
        MOVF    Old_RB,W        ;Send New RB value  
        MOVWF   xmt_byte
        BCF     T_B,RB_NEW_B    ;Clear need to send out newest value 
        BSF     T_B,OState_B    ;Start Task #1 & Lock Out Others
        GOTO    T4_S
_0058   BTFSS   T_B,S_5_S_B     ;Serial In 5 secs of inactivity
        GOTO    Task_5
        MOVLW   H'FF'           ;Tell of inactivity of Serial In 
        MOVWF   xmt_byte
        BCF     T_B,S_5_S_B     ;Clear need to send msg  
        BSF     T_B,OState_B    ;Start Task #1 & Lock Out Others
T4_S                            ;Start Task #1
        CALL    Do_OState

Task_5          ;Task #5 - Monitor Level Reset Input Line - Always Running ! 
        BTFSC   PORTA,Level_Reset
        GOTO    Task_6
        MOVLW   LED_OFF_MODE    ;Lowest Level Indicator output
        MOVWF   LED_Mode

Task_6          ;Task #6 - Debounce 8 bit Input Sensors - Runs every 20 mS 
        INCF    T_20_mS_CO      ;Inc Counter - Time Unit = 512 uS
        MOVLW   H'27'           ;Used to debounce the input 
        SUBWF   T_20_mS_CO,W
        BTFSS   STATUS,Z
        GOTO    _0065
        CLRF    T_20_mS_CO      ;Reset T_20_mS_CO to start over again
        MOVF    PORTB,W         ;Last copy of RB same as Current ? 
        SUBWF   Last_RB,W
        BTFSC   STATUS,Z        
        GOTO    _0062
        MOVF    PORTB,W         ;Store Current RB - diff from Last
        MOVWF   Last_RB
        GOTO    _0063
_0062   MOVF    Last_RB,W       ;New Old RB <- same value over 20 mS 
        MOVWF   Old_RB
_0063   MOVF    Old_RB          ;See if RB is now 0 
        BTFSC   STATUS,Z        ;RB == 0 ? then keep timer running 
        GOTO    _0064
        CLRF    T_5_M_LO        ;Reset 5 Min Timer 
        CLRF    T_5_M_HI        ;  still not zero yet 
_0064   CALL    D_H_E_L         ;Determine the Highest Error Level
        BTFSS   T_B,RB_NEW_B    ;Check for Lost Data Error 
        BCF     FLAGS,RB_Overflow
        BTFSC   T_B,RB_NEW_B
        BSF     FLAGS,RB_Overflow
        BSF     T_B,RB_NEW_B    ;Every 20 mS send Old_RB out
                
                ;Heart Beat - Time unit = 131072 uS for Tasks #7, #8 & #9
_0065   
        BSF     STATUS,RP0      ;Point to BANK 1
        MOVLW   RA_TRIS         ;RA TRIS - refresh 
        MOVWF   TRISA
        MOVLW   H'FF'           ;RB TRIS - refresh 
        MOVWF   TRISB
        BCF     STATUS,RP0      ;Point to BANK 0
        DECFSZ  cc              ;Step-up time units * 256 
        GOTO    Inc_Time

Task_7          ;Task 7 - Output Highest Level Indication on LED
        BTFSS   LED_Mode,LED_B  ;Is LED active ? 
        GOTO    Task_8
        CALL    Do_LED          ;Handle LED timing
                                  
Task_8          ;Task #8 - 5 Second Serial Input Lack of Activity Timer
        BTFSS   T_B,T_5_S_B     ;5 Sec Timer Active ?  
        GOTO    Task_9
        INCF    T_5_S_CO        ;Inc Counter - Time Unit = 131072 uS 
        MOVLW   H'26'           ;Check T_5_S_CO if time  
        SUBWF   T_5_S_CO,W
        BTFSS   STATUS,Z
        GOTO    Task_9
        CLRF    T_5_S_CO        ;Reset T_5_S_CO 
        MOVLW   LED_ON_MODE     ;Highest Level Indicator output 
        MOVWF   LED_Mode
        BTFSS   T_B,S_5_S_B     ;Check if Lost Data Error 
        BCF     FLAGS,S_5_S_Overflow
        BTFSC   T_B,S_5_S_B
        BSF     FLAGS,S_5_S_Overflow
        BSF     T_B,S_5_S_B     ;Send notice of 5 seconds of inaction  

Task_9          ;Task #9 - 5 Min. Lack of Severe Error from Sensors Reset Timer 
        BTFSS   T_B,T_5_M_B     ;5 Min Timer Active ? 
        GOTO    Task_A
        INCF    T_5_M_LO        ;Inc LO Counter; Time Unit = 131072 uS 
        BTFSC   STATUS,Z        ;See if carry needs to be passed on ?
        INCF    T_5_M_HI        ;Inc HI Counter; Time Unit = 131072 uS
        MOVLW   H'08'           ;#2288<  Check T_5_M_HI if time  
        SUBWF   T_5_M_HI,W
        BTFSS   STATUS,Z
        GOTO    Task_A
        MOVLW   H'F0'           ;#2288>  Check T_5_M_LO if time 
        SUBWF   T_5_M_LO,W
        BTFSS   STATUS,Z
        GOTO    Task_A
        CLRF    T_5_M_LO        ;Reset T_5_M_LO 
        CLRF    T_5_M_HI        ;Reset T_5_M_HI 
        MOVLW   LED_OFF_MODE    ;Lowest Level Indicator output 
        MOVWF   LED_Mode
Task_A   
        GOTO    Inc_Time        ;Loop Forever

;****           Determine the Highest Error Level & Start Task #7
D_H_E_L         
        MOVLW   H'07'           ;Check top 7 bits 
        MOVWF   temp
        MOVF    Old_RB,W        ;Get copy of 7 debounced Sensor Input
        MOVWF   tmp
_0070   RLF     tmp             ;Put top bit into C bit
        BTFSC   STATUS,C        ;Check if C bit is set 
        GOTO    _0072
        DECFSZ  temp            ;Continue to check lesser bits
        GOTO    _0070
_0071   MOVF    PORTB,W         ;Restore current value of RB 
        MOVWF   tmp
        RETURN
_0072   MOVF    LED_Mode,W      ;Get current Level Indicator 
        ANDLW   H'07'           ;Get only      "       "  
        MOVWF   tmp             ;Store into a tempary register 
        MOVF    temp,W          ;Check if already at this Level
        SUBWF   tmp,W
        BTFSC   STATUS,C
        GOTO    _0071
        MOVLW   H'88'           ;Start to build LED_Mode 
        IORWF   temp,W          ;Put new Level Indicator into reg 
        MOVWF   LED_Mode        ;Store new LED Mode 
        GOTO    _0071

;******         Clear Registers 20-7Fh, A0-C0  
Clear_Regs   
        MOVLW   H'7F'           ;First regs to clear in Bank 0
        MOVWF   FSR
Loop_C1 CLRF    INDF            ;Clear reg 
        DECF    FSR             ;point to next reg to clear 
        MOVLW   H'20'           ;Dec temp, jump if not done
        SUBWF   FSR,W
        BTFSC   STATUS,C
        GOTO    Loop_C1
        
        MOVLW   H'C0'           ;First regs to clear in Bank 1
        MOVWF   FSR
Loop_C2 CLRF    INDF            ;Clear reg 
        DECF    FSR             ;point to next reg to clear 
        MOVLW   H'A0'           ;Dec temp, jump if not done
        SUBWF   FSR,W
        BTFSC   STATUS,C
        GOTO    Loop_C2
        CLRF    FSR             ;Lastly clear FSR reg 
        RETURN

;******         TMR0 IRS - Set Time_Bit for background tasks
Do_Inc_Time                     
        BSF     FLAGS,Time_Bit  ;Tell background tasks of overflow
        BCF     INTCON,RTIF     ;Clear for next overflow
        RETURN

;******         Handle Interrupts Here
Interrupt
PUSH:   MOVWF   Temp_W
        SWAPF   STATUS,W
        MOVWF   Temp_Stat
        MOVF    FSR,W
        MOVWF   Temp_FSR
        BCF     STATUS,RP0      ;Point to BANK 0 - Very IMPORTANT !!!!!
        
        BTFSC   PIR1,TMR2IF
        CALL    Do_OState       ;Go Do Task #1 - all states 
        
        BTFSC   T_B,IState1_B   ;INTF will set even if INTE is cleared
        GOTO    I1
        BTFSC   INTCON,INTF
        CALL    Do_I1State      ;Go Do Task #2 - 0 state only
I1:      
        BTFSC   PIR1,TMR1IF
        CALL    Do_I1State      ;Go Do Task #2 - 1-B states

        BTFSC   INTCON,RTIF
        CALL    Do_Inc_Time     ;Go Inc Time_Bit every 512uS

POP:    MOVF    Temp_FSR,W
        MOVWF   FSR
        SWAPF   Temp_Stat,W
        MOVWF   STATUS
        SWAPF   Temp_W
        SWAPF   Temp_W,W
        RETFIE                  ;Return from Interrupt

        END

